#ifndef BOUNDING_BOX_H
#define BOUNDING_BOX_H

#include <nmath/vec3.h>
#include <nmath/vec2.h>

#include <vector>

using namespace std;
using namespace nacb;

class BoundingBox {
public:
  BoundingBox() {
    minVert = Vec3f(0, 0, 0);
    maxVert = Vec3f(0, 0, 0);
  }

  BoundingBox(const std::vector<Vec3f> & vert) {
    maxVert = Vec3f(-10e10, -10e10, -10e10);
    minVert = Vec3f( 10e10,  10e10,  10e10);    

    for (int i=0; i<vert.size(); i++) {
      minVert = minVert.min(vert[i]);
      maxVert = maxVert.max(vert[i]);      
    }
  }

  Vec2f getBoundsForMatrix(const Matrix & P) {
    float minz = 1e10, maxz = -1e10;

    // Test out all 8 sides of the cube.
    for (int i=0; i<8; i++) {
      Vec3f vert((i & 0x1) ? minVert.x : maxVert.x,
		 (i & 0x2) ? minVert.y : maxVert.y,
		 (i & 0x4) ? minVert.z : maxVert.z);
      float zout;
      project(P, vert, &zout);
      
      minz = std::min(minz, zout);
      maxz = std::max(maxz, zout);
    }
    return Vec2f(minz, maxz);
  }

  Vec3f minVert;
  Vec3f maxVert;
};

#endif
