#define GL_GLEXT_PROTOTYPES


#include <vector>
#include <map>

#include <utigl/glwindow.h>
#include <utigl/glcommon.h>
#include <utigl/drawcamera.h>

#include <nimage/image.h>
#include <ik/mesh.h>
#include <stereo/simplesequence.h>
#include <autorecon/recon_globals.h>
#include <nmisc/timer.h>
#include <set>
#include <utigl/fbo.h>
#include "tracker.h"

using namespace nacb;

class Win : public GLWindow {
public:
  Tracker * tracker;
  bool tracking;
  bool imageBased;

  Win(std::vector<std::string> seqNames, 
      const char * meshName, const char * texName) : GLWindow(640, 480) {
    tracker = new Tracker(seqNames, meshName, texName);

    glEnable(GL_DEPTH_TEST);

    imageBased = false;
    tracking = false;
  }

  ~Win() {
    delete tracker;
  }

  void drawScene() {
    tracker->draw();
    
    // Use the display loop instead of a thread (due to openGL thread sharing).
    if (tracking) {
      if (!tracker->trackerLoopIteration(imageBased)) {
	tracking = false;
	printf("Tracking complete.\n");
	setRefreshRate(0);
      }
      TimeChart::instance().print();
    }
  }

  bool keyboard(unsigned char c, int x, int y) {
    switch (c) {
    case 'v':
      tracker->updateVisibility();
      break;

    case 'n':
      tracker->imagei++;
      tracker->loadImages();
      break;
      
    case 't':
      tracker->textureTrack();
      break;

    case 'I':
      imageBased = !imageBased;
      break;

    case 'T':
      {
	tracking = !tracking;
	if (tracking)
	  setRefreshRate(50);
	else
	  setRefreshRate(0);
      }
      break;

    case ' ':
      tracker->imageTrack();
      break;
    }
    refresh();
    return true;
  }
};


int main(int ac, char * av[]){
  if (ac <= 3) {
    printf("Need the sequence as an argument.\n");
    return 0;
  }
  
  //test_derivatives(av[1], av[2], av[3]);
  vector<std::string> seqNames;
  for (int i=3; i<ac; i++) {
    seqNames.push_back(av[i]);
  }

  Win * win = new Win(seqNames, av[1], av[2]);
  win->loop(1);
  return 0;
}
