/**
   The original computation of the derivatives.  This was meant
   to mostly be throwaway code.

   I was first using a bunch of precalibrated images to test the derivative
   in different views (when using one of them as a reference view).
   
   There are a number of flags (recenter, and move) that were used to 
   try out different configurations.  This seemed to work, but the code
   contained quite a bit of debug code (for drawing in OpenGL, etc.).  I pulled
   out most of this and put it in a new, cleaner location.
      

   Neil Birkbeck Jan, 2010.
*/

// Compare the derivatives when computed analytically using images,
// to when they are computed using a texture space.
void check_derivatives(Mesh & mesh, 
		       seq_t & seq0,
		       seq_t & seq, 
		       const nacb::Imagef & texture,
		       int drawk = 0) {
  double eps = 1e-5;
  assert(seq0.load(0));


  Matrix refK = seq0.A;
  Matrix refE = seq0.E;
  Matrix refP = seq0.P;

  const bool recenter = false;
  
  if(recenter) {    
    for (int i=0; i<mesh.vert.size(); i++) {
      mesh.vert[i] = affine3D(seq0.E, mesh.vert[i]);
    }
    for (int i=0; i<mesh.tvert.size(); i++) {
      mesh.tvert[i].z = 1;
    }
  }
  refE = Matrix::eye(4, 4);
  refP = refK*Matrix::eye(3, 4)*refE;    
  
  Imagef Tgrad = texture.gradient(0);

  // Current transformation parameters.
  Matrix oE = Matrix::eye(4, 4);
  bool move = true; // Test out having different cameras positioned.

  if (move) {
    oE = Matrix::trans(0.1, -0.1, 0.1)*Matrix::rotz(0.3)*Matrix::rotx(0.1)*Matrix::roty(0.1);
    Matrix oEinv = oE.inverse();
    for (int i=0; i<mesh.vert.size(); i++) {
      mesh.vert[i] = affine3D(oEinv, mesh.vert[i]);
    }
    for (int i=0; i<mesh.tvert.size(); i++) {
      mesh.tvert[i].z = 1;
    }
  }
  
  

  for (int i=1; i<=1; i++) {
    if (seq.load(i)) {
      Imagef gout_image(512, 512, 1);
      Imagef gout_tex(512, 512, 1);
      Imagef grad = seq.image.gradient(0);

      gout_image = 0;
      gout_tex = 0;

      if (recenter) {
	seq.E = seq.E*seq0.E.inverse();
	seq.P = seq.A*Matrix::eye(3, 4)*seq.E;
	
	seq.P.printMatlab("P");
      }
      Matrix curParms = extract_parms(seq.E);
      curParms.printMatlab("curParms");
      vector<Matrix> Eup(6, Matrix::zeros(4, 4));
      vector<Matrix> matup(6, Matrix::zeros(6, 1));

      if (move) {
	curParms = extract_parms(oE);

	for (int pi=0; pi<6; pi++) {
	  printf("on %d\n", pi);
	  Matrix parms = curParms.copy();
	  parms[pi] += eps;
	  matup[pi] = seq.A*Matrix::eye(3, 4)*seq.E*parms_to_matrix(parms);
	  Eup[pi] = seq.E*parms_to_matrix(parms);
	}
      }
      else {
	for (int pi=0; pi<6; pi++) {
	  printf("on %d\n", pi);
	  Matrix parms = curParms.copy();
	  parms[pi] += eps;
	  matup[pi] = seq.A*Matrix::eye(3, 4)*parms_to_matrix(parms);
	  Eup[pi] = parms_to_matrix(parms);
	}
      }

      for (int t=0; t<mesh.tris.size(); t++) {
	Vec3f e1 = mesh.vert[mesh.tris[t].vi[1]] - mesh.vert[mesh.tris[t].vi[0]];
	Vec3f e2 = mesh.vert[mesh.tris[t].vi[2]] - mesh.vert[mesh.tris[t].vi[0]];
	Vec3f c = mesh.vert[mesh.tris[t].vi[0]];
	Vec3f n = e1.cross(e2);
	double d = c.dot(n);
	Matrix nd = (Matrix)(n*(1.0/d));

	Matrix eye = refP;//seq0.A*Matrix::eye(3, 4);//Matrix::eye(3, 4);
	Matrix H0 = linear_homography(eye, mesh, t);
	Matrix Euse = seq.E*oE;
	Matrix H1 = linear_homography(seq.A, Euse, c, n);

	Matrix Tinv = (H1*H0).inverse();

	
	// Another sanity check
	for (int k=0; k<3; k++) {
	  Vec2f cc = project(seq.P*oE, mesh.vert[mesh.tris[t].vi[k]]);
	  Matrix T = H1*H0;
	  Vec2f Tinvc = Hwarp(T, mesh.tvert[mesh.tris[t].tci[k]]);
	  
	  printf("D: %f %f  %f %f\n", cc.x, cc.y, Tinvc.x, Tinvc.y);
	}

	// Derivatives of the F matrix.
	vector<Matrix> Fderiv(6, Matrix::zeros(6, 1));
	for (int pi=0; pi<6; pi++) {
	  Matrix Up = linear_homography(seq.A, Eup[pi], c, n);
	  Fderiv[pi] = ((Up*H0 - H1*H0)*(1.0/eps));
	  Fderiv[pi].printMatlab("Fd");
	}
	H1.printMatlab("H1");

	for (double a=0.02; a<=0.98; a+=1.0/192) {
	  for (double b=0.02; b<=0.98-a; b+=1.0/192) {
	    double c = 1.0 - a - b;
	    Vec3f point = mesh.vert[mesh.tris[t].vi[0]]*a
	      + mesh.vert[mesh.tris[t].vi[1]]*b
	      + mesh.vert[mesh.tris[t].vi[2]]*c;
	    Vec3f tc = mesh.tvert[mesh.tris[t].tci[0]]*a
	      + mesh.tvert[mesh.tris[t].tci[1]]*b
	      + mesh.tvert[mesh.tris[t].tci[2]]*c;

	    Vec3f torig = tc;
	    Vec3f tcn = tc;

	    tc.x *= Tgrad.w;
	    tc.y *= Tgrad.h;

	    // Standard derivative
	    for (int k=drawk; k<=drawk; k++) {
	      float zout = 0;
	      Vec3f ptrans = affine3D(oE, point);
	      Vec2f proj = project(seq.P, ptrans, &zout);
	      Vec2f pup = project(matup[k], point);
	      
	      Vec2f deriv = (pup - proj)*(1.0/eps);
	      Vec2f imd;
	      grad.bilinear(proj.x, proj.y, imd.data);
	      
	      float res = imd.dot(deriv);
	      
	      if (k == drawk) {
		printf("drawing: %f %f\n", proj.x, proj.y);
		glBegin(GL_POINTS);
		glVertex2f(proj.x, proj.y);
		glEnd();

		glBegin(GL_LINES);
		glVertex2f(proj.x, proj.y);
		glVertex2f(proj.x + deriv.x/10, proj.y + deriv.y/10);
		glEnd();
	      }

	      // The point derivatives are 
	      Vec2f Timd;
	      Tgrad.bilinear(round(tc.x), round(tc.y), Timd.data);

	      Matrix Pd = pd_matrix(torig); 
	      Matrix F = (Fderiv[k]*((Matrix)torig));
	      
	      // Current estimate of rotation parameters.
	      Matrix Et = seq.E*oE;
	      Matrix Rt = Et.submatrix(0, 0, 3, 3);
	      Matrix t = Et.submatrix(0, 3, 3, 1);

	      // The analytical derivatives will look like below.
	      if (k < 3) {
		Matrix Rderiv = (Eup[k] - seq.E*oE).submatrix(0, 0, 3, 3)*(1.0/eps);
		F = seq.A*Rderiv*seq.A.inverse()*H0*((Matrix)torig);
		F.printMatlab("F");
		// F =(Fderiv[k]*((Matrix)torig));

		Matrix Ainv = seq.A.inverse();
		Matrix AinvTrans = Ainv.transpose();
		Matrix RtInvTrans = Rt.inverse().transpose();
		Matrix RderivTrans = Rderiv.transpose();
		Matrix u = ((Matrix)torig);
		Matrix Ss[3] = {kronecker_product(Matrix::eye(3, 3), (H0*u).transpose()),
				kronecker_product(Matrix::eye(3, 3), nd.transpose()) * kronecker_product(Matrix::eye(9, 9), (H0*u).transpose()),
				kronecker_product(Matrix::eye(3, 3), nd.transpose()) * kronecker_product(Matrix::eye(9, 9), (H0*u).transpose())};
		Matrix Ms[3] = {vectorize(AinvTrans*RderivTrans*RtInvTrans),
				vectorize(AinvTrans*RderivTrans*RtInvTrans*kronecker_product(Matrix::eye(3, 3), RtInvTrans*t).transpose()),
				(vectorize(AinvTrans*RderivTrans*RtInvTrans*kronecker_product((Rt.inverse()*t), Matrix::eye(3, 3)).transpose()))*-1};

		Matrix sum = Matrix::zeros(3, 1);
		double dval = (1+(t.transpose()*Rt.inverse()*((Matrix)nd))[0]);
		for (int h=0; h<3; h++) {
		  Matrix p = Ss[h] * Ms[h];
		  sum += (p) * (1.0/dval);//.printMatlab("h");
		}
		

		double val = (t.transpose()*Rt.inverse()*nd)[0];
		Matrix nofacts[3] = {Rt.inverse()*Rderiv*Ainv*H0*u,
				     Matrix::eye(3, 3)*val*Rt.inverse()*Rderiv*Ainv*H0*u,
				     Rt.inverse()*t*nd.transpose()*Rt.inverse()*Rderiv*Ainv*H0*u*-1};
		
		Matrix nofact = Matrix::zeros(3, 1);
		for (int h=0; h<3; h++){
		  nofact += nofacts[h];
		}
		nofact *= (1.0/dval);

		static int cnt = 0;
		cnt++;

		if (cnt % 10 == 0) {
		  printf("termk-0\n");
		  double factor = 100.0;
		  for (int k=0; k<3; k++) {
		    ((Ss[k] * Ms[k])*factor).printMatlab("termk");
		    (nofacts[k]*factor).printMatlab("termk-check");
		  }

		  seq.A*Rderiv*seq.A.inverse();
		  sum.printMatlab("sum");
		  nofact.printMatlab("sum-check");
		 
		  (Pd * Tinv * F).printMatlab("sep-check");		
		  (Pd * H0.inverse()*seq.A*nofact).printMatlab("sep-nofact");
		  (Pd * H0.inverse()*seq.A*sum).printMatlab("sep");
		}
	      }
	      else {
		(((Matrix)n).transpose()).printMatlab("n");
		Matrix rhs = (H0*((Matrix)torig)).transpose();
		double z = (nd.transpose()*seq.A.inverse()).dot(rhs);
		Matrix row = Matrix::zeros(3, 1);
		row[k-3] = 1;
		F = seq.A*row*z;
		F.printMatlab("F");
		

		Matrix Ss[3] = {Matrix::eye(3, 3)*z, 
				kronecker_product(Matrix::eye(3, 3)*z, ((Matrix)nd).transpose()),
				kronecker_product(Matrix::eye(3, 3)*z, ((Matrix)nd).transpose())};
		Matrix Ms[3] = {Rt.inverse(), 
				kronecker_product(Matrix::eye(3, 3), Rt.inverse().transpose()*t)*Rt.inverse(),
				kronecker_product(Rt.inverse()*t, Matrix::eye(3, 3)) * Rt.inverse() * -1};
		
		double dval = (1+(t.transpose()*Rt.inverse()*((Matrix)nd))[0]);
		double dval0 = ((t.transpose()*Rt.inverse()*((Matrix)nd))[0]);
		Matrix nofact = (Matrix::eye(3, 3) + Matrix::eye(3, 3)*dval0 - Rt.inverse()*t*nd.transpose())*Rt.inverse()*(Matrix::eye(3, 3)*z);
		nofact *= (1.0/dval); // This undoes the normalization, can't see how this scalar could be a problem, as it can multiply the M, matrix.  This scale factor can cause derivatives do differ.

		Matrix nofacts[3] = {Rt.inverse()*Matrix::eye(3, 3)*z,
				     Matrix::eye(3, 3)*dval0*Rt.inverse()*Matrix::eye(3, 3)*z,
				     (Rt.inverse()*t*nd.transpose())*Rt.inverse()*Matrix::eye(3,3)*(-z)};

		Matrix sum = Matrix::zeros(3, 3);
		for (int h=0; h<3; h++) {
		  sum += (Ss[h] * Ms[h]) * (1.0/dval);//.printMatlab("h");
		}
				     
//nofact = (Rt.inverse()*dval - Rt.inverse()*t*(nd.transpose())*Rt.inverse())*(Matrix::eye(3, 3)*z);
		
		static int cnt = 0;
		cnt++;
		if (cnt % 10 == 0) {
		  printf("termk-0\n");

		  for (int k=0; k<3; k++) {
		    (Ss[k] * Ms[k]).printMatlab("termk");
		    nofacts[k].printMatlab("termk-check");
		  }

		  sum.printMatlab("sum");
		  nofact.printMatlab("sum-check");
		 
		  (Pd * Tinv * F).printMatlab("sep-check");		
		  (Pd * H0.inverse()*seq.A*nofact).printMatlab("sep-nofact");
		  (Pd * H0.inverse()*seq.A*sum).printMatlab("sep");
		}		
		//F =(Fderiv[k]*((Matrix)torig));
		//F.printMatlab("Fd");
	      }

	      Matrix Jt = (Pd * Tinv * F); //Fderiv[k]);;
	      //J.printMatlab("J");

	      if (k == drawk) {
		glBegin(GL_LINES);
		glVertex2f(800 + torig.x*600, torig.y*600);
		glVertex2f(800 + (torig.x + Jt[0]/10)*600, (torig.y + Jt[1]/10)*600);
		glEnd();
	      }

	      Jt *= 512;
	      double tres = Timd.dot(Jt);
	      printf("compare i: %f  %f   %f\n", res, tres, res/tres);

	      /*
	      printf("colors: %f %f\n", 
		     seq.image.bilinear(proj.x, proj.y, 0),
		     texture.bilinear(tc.x, tc.y, 0));
	      */
	      for (int h=0; h<3; h++) {
		seq.image((int)proj.x, (int)proj.y, h) = texture.bilinear(tc.x, tc.y, h);
	      }
	      gout_image(tc.x, tc.y) = res;
	      gout_tex(tc.x, tc.y) = tres;
	      
	      //seq.image((int)proj.x, (int)proj.y, 3) = 0;
	    }
	  }
	}
      }
      double n, x;
      gout_tex.getRange(n, x);
      ((gout_tex - n)*(1.0/(x-n))).save("/tmp/gout_tex.png");
      ((gout_image - n)*(1.0/(x-n))).save("/tmp/gout_image.png");
      
      seq.image.save("/tmp/image.png");
    }
    else {
      printf("Cannot load %d\n", i);
      break;
    }
  }
}



class Win: public GLWindow {
public:
  seq_t seq0;
  seq_t seq;
  Mesh mesh;
  std::string fname;
  nacb::Imagef image;

  Win(const char * _fname) : GLWindow(1200, 600) {
    drawk = 0;
    fname = _fname;
  }

  
  void applyProjection(){
    glOrtho(0, 1200, 0, 600, -100, 100);
  }
  

  void drawQuad(double w, double h){
    glBegin(GL_QUADS);
     glTexCoord2f(0, 0);
     glVertex2f(0, 0);

     glTexCoord2f(1, 0);
     glVertex2f(w, 0);    

     glTexCoord2f(1, 1);
     glVertex2f(w, h);    

     glTexCoord2f(0, 1);
     glVertex2f(0, h);    
    glEnd();
  }
  
  void drawScene() {
    seq0 = seq_t(fname);
    seq = seq_t(fname);
    
    mesh.loadobj("/tmp/rigidtrack2.obj");
    image.load("/tmp/rigidtrack2.png");
    image = image.flip();

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);

    glPushMatrix();
    glScalef(12.0/14.0, 12./14., 1.0);

    glBindTexture(GL_TEXTURE_2D, 101);
    seq.load(1);
    seq.image.initTexture();
    glEnable(GL_TEXTURE_2D);
    drawQuad(800, 600);

    glPushMatrix();
    glTranslatef(800, 0, 0);    
    image.initTexture();
    drawQuad(600, 600);
    glPopMatrix();

    check_derivatives(mesh, seq0, seq, image, drawk);    
    glPopMatrix();
  }

  bool keyboard(unsigned char c, int x, int y) {
    switch (c) {
    case ' ':
      drawk = (drawk + 1)%6;
      break;
    }
  }

  int drawk;
};
