#include "jacobian.h"
#include <nimage/image.h>
#include <ik/mesh.h>
#include <stereo/simplesequence.h>
#include <autorecon/recon_globals.h>
#include <nmisc/timer.h>

#include "utils.h"

using namespace nacb;
typedef SimpleSequence<float> seq_t;


PointCache::PointCache(const Mesh & mesh,
		       int ti, 
		       const Vec3f & bary,
		       const Matrix & H0,
		       const seq_t & seq,
		       const Imagef & texture,
		       const std::vector<Imagef> & Tgrad) {
  int nchannels = Tgrad.size();

  point = mesh.vert[mesh.tris[ti].vi[0]]*bary.x
    + mesh.vert[mesh.tris[ti].vi[1]]*bary.y
    + mesh.vert[mesh.tris[ti].vi[2]]*bary.z;

  Vec3f tc = mesh.tvert[mesh.tris[ti].tci[0]]*bary.x
    + mesh.tvert[mesh.tris[ti].tci[1]]*bary.y
    + mesh.tvert[mesh.tris[ti].tci[2]]*bary.z;

  Vec3f e1 = mesh.vert[mesh.tris[ti].vi[1]] - mesh.vert[mesh.tris[ti].vi[0]];
  Vec3f e2 = mesh.vert[mesh.tris[ti].vi[2]] - mesh.vert[mesh.tris[ti].vi[0]];
  Vec3f c = mesh.vert[mesh.tris[ti].vi[0]];
  n = e1.cross(e2);
  n.normalize();

  d = c.dot(n);
  nd = (Matrix)(n*(1.0/d));

  Vec3f torig = tc;
  Vec3f tcn = tc;
	
  tc.x *= Tgrad[0].w;
  tc.y *= Tgrad[0].h;

  texture.bilinear(tc.x, tc.y, textureColor.data);

  Matrix Pd = pd_matrix(torig); 

  normalization = 1;

    
  Matrix u = ((Matrix)torig);

  Matrix rhs = (H0*((Matrix)torig)).transpose();
  double z = (nd.transpose()*seq.A.inverse()).dot(rhs);


  Matrix gradChannels(nchannels, 2);

  for (int h=0; h<nchannels; h++) {
    Vec2f Timd;
    Tgrad[h].bilinear(tc.x, tc.y, Timd.data);
      
    gradChannels(h, 0) = Timd.x;
    gradChannels(h, 1) = Timd.y;		      
  }
    
  Matrix premul = gradChannels * Pd * H0.inverse() * seq.A;

  for (int k=0; k<3; k++) {    
    Matrix Sk, Mk;
    Matrix sum;

    Sk = premul * Matrix::cat(1, kronecker_product(Matrix::eye(3, 3), (H0*u).transpose()),
			      Matrix::cat(1, kronecker_product(Matrix::eye(3, 3), nd.transpose()) * kronecker_product(Matrix::eye(9, 9), (H0*u).transpose()),
					  kronecker_product(Matrix::eye(3, 3), nd.transpose()) * kronecker_product(Matrix::eye(9, 9), (H0*u).transpose())));
      
    if (k == 0) {
      S = Sk.copy();
    }
    else {
      S = Matrix::cat(1, S, Sk);
    }
  }
    
  Matrix Sk;

  Sk = premul * Matrix::cat(1, Matrix::eye(3, 3)*z, 
			    Matrix::cat(1, kronecker_product(Matrix::eye(3, 3)*z, ((Matrix)nd).transpose()),
					kronecker_product(Matrix::eye(3, 3)*z, ((Matrix)nd).transpose())));

  //S = S.getRow(k - 3);
  //M = M.getColumn(k - 3);		
  //Matrix Rsub = seq.E.submatrix(0, 0, 3, 3)
  // sum = sum * Rsub; // Rsub needed to change coordinates.

  S = Matrix::cat(1, S, Sk);
  Strans = S.transpose();
  L = Strans*S;

  int nz = 0;
  for (int i=0; i<L.m*L.n; i++) {
    nz += fabs(L.data[i])<1e-6;
  }
  //sum = sum.getColumn(k - 3);
  //S.printMatlab("S");

  A = Matrix::zeros(1, 1);
}


void PointCache::updateA(seq_t & seq, Matrix & P, Matrix & Mtrans) {
  Vec2f pt = project(P, point); 
  Vec3f imageColor;

  seq.image.bilinear(pt.x, pt.y, imageColor.data);

  Matrix E = (imageColor - textureColor);
  E *= (1.0/normalization);
  //A = Mtrans * (Strans * E);
  //A.printMatlab("A");

  Matrix T = Strans*E;
  int sizes[4] = {0, 63, 63*2, 63*3};
    
  A = Matrix(6, 1);
  for (int j=0; j<3; j++) {
    double res = 0;
    for (int i=0; i<63; i++) {
      res += Mtrans(j, sizes[j] + i) * T.data[i + sizes[j]];
    }
    A[j] = res;
  }

  for (int j=0; j<3; j++) {
    double res = 0;
    for (int i=0; i<21; i++) {
      res += Mtrans(3 + j, sizes[3] + i) * T.data[i + sizes[3]];
    }
    A[j + 3] = res;
  }
}


Matrix buildM(const seq_t & seq, 
	      const Matrix & Rt,
	      const Matrix & t, 
	      vector<Matrix> & Ederiv){
  Matrix Ainv = seq.A.inverse();
  Matrix AinvTrans = Ainv.transpose();
  Matrix RtInvTrans = Rt.inverse().transpose();

  Matrix M = Matrix::zeros(63*3 + 21, 6);

  for (int k=0; k<3; k++) {
    Matrix Rderiv = Ederiv[k].submatrix(0, 0, 3, 3);
    Matrix RderivTrans = Rderiv.transpose();
    Matrix Mk;

    Mk = Matrix::cat(0, vectorize(AinvTrans*RderivTrans*RtInvTrans),
		     Matrix::cat(0, vectorize(AinvTrans*RderivTrans*RtInvTrans*kronecker_product(Matrix::eye(3, 3), RtInvTrans*t).transpose()),
				 (vectorize(AinvTrans*RderivTrans*RtInvTrans*kronecker_product((Rt.inverse()*t), Matrix::eye(3, 3)).transpose()))*-1));
    M.setSubmatrix(63*k, k, Mk);
  }
  Matrix Mt;
  Mt = Matrix::cat(0, Rt.inverse(), 
		   Matrix::cat(0, kronecker_product(Matrix::eye(3, 3), Rt.inverse().transpose()*t)*Rt.inverse(),
			       kronecker_product(Rt.inverse()*t, Matrix::eye(3, 3)) * Rt.inverse() * -1));

  Matrix Rsub = seq.E.submatrix(0, 0, 3, 3);
  Mt = Mt*Rsub;
  M.setSubmatrix(63*3, 3, Mt);
  // Ederiv[k] = (Eup[k] - seq.E*oE).submatrix(0, 0, 3, 3)*(1.0/eps);
  //sum = Sk * Mk * (1.0/dval);
  return M;
}



std::vector<Imagef> derivatives(const Mesh & mesh, 
				const seq_t & seq, 
				const nacb::Imagef & texture,
				Matrix & curParms, 
				bool textureSpace) {
  const double eps = 1e-5;
  Matrix refP = seq.A*Matrix::eye(3, 4);
  Matrix oE = parms_to_matrix(curParms);

  vector<Matrix> Eup(6, Matrix::zeros(4, 4));
  vector<Matrix> Ederiv(6, Matrix::zeros(4, 4));
  vector<Matrix> matup(6, Matrix::zeros(6, 1));
  int nparms = curParms.m*curParms.n;

  int nchannels = std::min(seq.image.nchannels, std::min(3, texture.nchannels));
  int minchan = 0;
  
  vector<Imagef> grad(nchannels, Imagef(1, 1, 1));
  vector<Imagef> Tgrad(nchannels, Imagef(1, 1, 1));

  for (int k=0; k<nchannels; k++) {
    grad[k] = seq.image.gradient(k + minchan);
    Tgrad[k] = texture.gradient(k + minchan);
    
    // FIXME: should really be scaling up the channels.
    Tgrad[k] *= Tgrad[k].w;  
  }
 
  vector<Imagef> tex(nparms);

  for (int pi=0; pi<curParms.m*curParms.n; pi++) {
    Matrix parms = curParms.copy();
    parms[pi] += eps;
    matup[pi] = seq.A*Matrix::eye(3, 4)*seq.E*parms_to_matrix(parms);
    Eup[pi] = seq.E*parms_to_matrix(parms);

    tex[pi] = Imagef(texture.w, texture.h, nchannels);
    tex[pi] = 0;

    Ederiv[pi] = (Eup[pi] - seq.E*oE)*(1.0/eps);
  }

  bool simpleTexDeriv = false;

  Matrix Pcomp = seq.A*Matrix::eye(3, 4)*seq.E*oE;

  // Current estimate of rotation parameters.
  Matrix Et = seq.E*oE;
  Matrix Rt = Et.submatrix(0, 0, 3, 3);
  Matrix t = Et.submatrix(0, 3, 3, 1);
  
  Matrix M = buildM(seq, Rt, t, Ederiv);

  for (int ti=0; ti<mesh.tris.size(); ti++) {
    Vec3f e1 = mesh.vert[mesh.tris[ti].vi[1]] - mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f e2 = mesh.vert[mesh.tris[ti].vi[2]] - mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f c = mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f n = e1.cross(e2);
    n.normalize();

    double d = c.dot(n);
    Matrix nd = (Matrix)(n*(1.0/d));
    
    Matrix H0 = linear_homography(refP, mesh, ti);
    Matrix H1 = linear_homography(seq.A, Et, c, n);
    
    Matrix Tinv = (H1*H0).inverse();
        
    // Derivatives of the F matrix.
    vector<Matrix> Fderiv(nparms, Matrix::zeros(6, 1));
    for (int pi=0; pi<nparms; pi++) {
      Matrix Up = linear_homography(seq.A, Eup[pi], c, n);
      Fderiv[pi] = ((Up*H0 - H1*H0)*(1.0/eps));
      //Fderiv[pi].printMatlab("Fd");
    }
    //H1.printMatlab("H1");

        
    for (double a=0.02; a<=0.98; a+=1.0/192) {
      for (double b=0.02; b<=0.98-a; b+=1.0/192) {
	double c = 1.0 - a - b;
	Vec3f point = mesh.vert[mesh.tris[ti].vi[0]]*a
	  + mesh.vert[mesh.tris[ti].vi[1]]*b
	  + mesh.vert[mesh.tris[ti].vi[2]]*c;
	Vec3f tc = mesh.tvert[mesh.tris[ti].tci[0]]*a
	  + mesh.tvert[mesh.tris[ti].tci[1]]*b
	  + mesh.tvert[mesh.tris[ti].tci[2]]*c;
	
	Vec3f torig = tc;
	Vec3f tcn = tc;
	
	tc.x *= texture.w;
	tc.y *= texture.h;
	
	
	if (textureSpace && !simpleTexDeriv) {
	  PointCache cache(mesh, ti, Vec3f(a, b, c), H0, seq, texture, Tgrad);
	  cache.updateNormalization(Rt, t);	  

	  Matrix deriv = (cache.S * M) * (1.0 / cache.normalization);

	  for (int k=0; k<6; k++) {
	    for (int h=0; h<nchannels; h++) {
	      tex[k]((int)tc.x, (int)tc.y, h) = deriv(h, k);
	    }
	  }
	} 
	else {

	  for (int k=0; k<nparms; k++) {
	    float res[nchannels];
	    
	    if (!textureSpace) {	    
	      float zout = 0;
	      Vec2f proj = project(Pcomp, point, &zout);
	      Vec2f pup = project(matup[k], point);
	      
	      Vec2f deriv = (pup - proj)*(1.0/eps);
	      Vec2f imd;
	      
	      for (int h=0; h<nchannels; h++) {
		grad[h].bilinear(proj.x, proj.y, imd.data);
		
		res[h] = imd.dot(deriv);	    
	      }
	    }
	    else if (simpleTexDeriv) {
	      Matrix Jt;
	      Matrix Pd = pd_matrix(torig); 
	      
	      Matrix F = (Fderiv[k]*((Matrix)torig));	    
	      Jt = (Pd * Tinv * F);	    
	      
	      for (int h=0; h<nchannels; h++) {
		Vec2f Timd;
		Tgrad[h].bilinear(round(tc.x), round(tc.y), Timd.data);
		
		res[h] = Timd.dot(Jt);
	      }
	    }	  
	    for (int h=0; h<nchannels; h++) {
	      tex[k]((int)tc.x, (int)tc.y, h) = res[h];
	    }
	  }
	}
      }
    }
  }
  return tex;
}


/**
   Sample a bunch of points on the triangles of the mesh.
*/
std::vector<SamplePoint> extract_sample_points(Mesh & mesh, Imagef & texture, 
					       std::vector<Imagef> & Tgrad,
					       const vector<seq_t> & seqs) {
  std::vector<SamplePoint> points;
  fprintf(stderr, "Getting sample points.\n");
  
  for (int ti=0; ti<mesh.tris.size(); ti++) {
    printf("Sampling triangle %d\n", ti, points.size());

    Vec3f e1 = mesh.vert[mesh.tris[ti].vi[1]] - mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f e2 = mesh.vert[mesh.tris[ti].vi[2]] - mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f c = mesh.vert[mesh.tris[ti].vi[0]];
    Vec3f n = e1.cross(e2);
    n.normalize();

    double d = c.dot(n);
    Matrix nd = (Matrix)(n*(1.0/d));
    double sampleSpacing = 65 / seqs.size();
    
    for (double a=0.05; a<=0.95; a+=1.0/sampleSpacing) {
      for (double b=0.05; b<=0.95-a; b+=1.0/sampleSpacing) {
	double c = 1.0 - a - b;
	Vec3f point = mesh.vert[mesh.tris[ti].vi[0]]*a
	  + mesh.vert[mesh.tris[ti].vi[1]]*b
	  + mesh.vert[mesh.tris[ti].vi[2]]*c;
	Vec3f tc = mesh.tvert[mesh.tris[ti].tci[0]]*a
	  + mesh.tvert[mesh.tris[ti].tci[1]]*b
	  + mesh.tvert[mesh.tris[ti].tci[2]]*c;
	Vec3f tcorig = tc;

	tc.x *= texture.w;
	tc.y *= texture.h;

	Vec3f color;
	texture.bilinear(tc.x, tc.y, color.data);
	
	SamplePoint sample;
	
	sample.tc = tcorig;
	sample.point = point;
	sample.n = n;
	sample.textureColor = color;
	
	for (int i=0; i<seqs.size(); i++) {
	  Matrix refP = seqs[i].A * Matrix::eye(3, 4);
	  Matrix H0 = linear_homography(refP, mesh, ti);
	  sample.caches.push_back(PointCache(mesh, ti, Vec3f(a, b, c), H0, seqs[i], texture, Tgrad));
	}
	//sample.H0 = H0;
	//sample.nd = nd;
       
	points.push_back(sample);
      }
    }
  }
  printf("There are %d points\n", points.size());
  return points;
}

Matrix image_derivative(std::vector<SamplePoint> & points, 
			std::vector<Imagef> & grad, 
			seq_t & seq, 
			const Matrix & curParms, 
			int nchannels, Matrix & diff,
			std::vector<int> * vis) {
  int top = vis ? vis->size() : points.size();
  if (diff.m != points.size()*nchannels || diff.n != 1)
    diff = Matrix(top*nchannels, 1);

  Matrix oE = parms_to_matrix(curParms);

  vector<Matrix> Eup(6, Matrix::zeros(4, 4));
  vector<Matrix> Ederiv(6, Matrix::zeros(4, 4));
  vector<Matrix> matup(6, Matrix::zeros(6, 1));
  int nparms = curParms.m*curParms.n;
  const double eps = 1e-5;

  for (int pi=0; pi<curParms.m*curParms.n; pi++) {
    Matrix parms = curParms.copy();
    parms[pi] += eps;
    matup[pi] = seq.A*Matrix::eye(3, 4)*seq.E*parms_to_matrix(parms);
    Eup[pi] = seq.E*parms_to_matrix(parms);

    Ederiv[pi] = (Eup[pi] - seq.E*oE)*(1.0/eps);
  }

  Matrix Pcomp = seq.A*Matrix::eye(3, 4)*seq.E*oE;
  
  Matrix J(top*nchannels, 6);
  
  for (int i=0; i<top; i++) {
    int pi = i;

    if (vis)
      pi = (*vis)[i];

    float zout = 0;
    Vec2f proj = project(Pcomp, points[pi].point, &zout);
    
    for (int k=0; k<6; k++) {
      Vec2f pup = project(matup[k], points[pi].point);
      
      Vec2f deriv = (pup - proj)*(1.0/eps);
      Vec2f imd;
      
      for (int h=0; h<nchannels; h++) {
	grad[h].bilinear(proj.x, proj.y, imd.data);
	
	J(i*nchannels + h, k) = imd.dot(deriv);
      }
    }
    
    Vec3f imageColor;
    seq.image.bilinear(proj.x, proj.y, imageColor.data);
    Vec3f colorDiff = imageColor - points[pi].textureColor;
    
    for (int h=0; h<nchannels; h++)
      diff[i*nchannels + h] = colorDiff.data[h];
    
  }
  return J;
}


/**
   Convert the Jacobian J into a texture image using the sample points.
 */
vector<Imagef> jacobian_to_images(Matrix & J, vector<SamplePoint> & points) {
  std::vector<Imagef> tex;
  int nchannels = J.m/points.size();
  for (int i=0; i<J.n; i++) {
    tex.push_back(Imagef(512, 512, nchannels));
    tex.back() = 0;
  }
  
  for (int i=0; i<points.size(); i++) {
    int x = (int)round(points[i].tc.x * 512);
    int y = (int)round(points[i].tc.y * 512);
    
    for (int j=0; j<J.n; j++) {
      for (int k=0; k<nchannels; k++) {
	tex[j](x, y, k) = J(i*nchannels + k, j);
      }
    }
  }
  return tex;
}



