#ifndef JACOBIAN_H
#define JACOBIAN_H

#include <vector>

#include <nmath/vec3.h>
#include <stereo/simplesequence.h>
#include <autorecon/recon_globals.h>

#include "utils.h"

using namespace std;
using namespace nacb;

typedef SimpleSequence<float> seq_t;


/**
   Cache all the information required for the factorized matrix
   approach.  Some of the information (e.g., point) and textureColor
   are redundant with the SamplePoint.
 */
class PointCache {
public:
  Matrix S, Strans;
  Matrix L;

  Matrix A;

  double normalization;
  Matrix nd;

  Vec3f  n;
  double d;

  Vec3f point, textureColor;

  PointCache(const Mesh & mesh,
	     int ti, 
	     const Vec3f & bary,
	     const Matrix & H0,
	     const seq_t & seq,
	     const Imagef & texture,
	     const std::vector<Imagef> & Tgrad);

  void updateNormalization(Matrix & R, Matrix & t) {
    normalization = (1+(t.transpose()*R.inverse()*((Matrix)nd))[0]);
  }

  void updateA(seq_t & seq, Matrix & P, Matrix & Mtrans);

  void accumulateL(Matrix & Laccum) {
    Laccum += L*(1.0/(normalization*normalization));
  }

  void deaccumulateL(Matrix & Laccum) {
    Laccum -= L*(1.0/(normalization*normalization));
  }

  void accumulateA(Matrix & Aaccum) {
    Aaccum += A;
  }
};

/**
   A SamplePoint holds the relevant 3D information, a texture color, and
   some cached information for the factorized matrix multiplication.
 */
class SamplePoint { 
public:
  Vec3f point;
  Vec3f textureColor;
  Vec3f tc;
  Vec3f n;
  vector<PointCache> caches;
};

Matrix buildM(const seq_t & seq, 
	      const Matrix & Rt,
	      const Matrix & t, 
	      vector<Matrix> & Ederiv);


std::vector<SamplePoint> extract_sample_points(Mesh & mesh, Imagef & texture, 
					       std::vector<Imagef> & Tgrad,
					       const vector<seq_t> & seqs);

Matrix image_derivative(std::vector<SamplePoint> & points, 
			std::vector<Imagef> & grad, 
			seq_t & seq, 
			const Matrix & curParms, 
			int nchannels, Matrix & diff, std::vector<int> * vis = 0);

vector<Imagef> jacobian_to_images(Matrix & J, vector<SamplePoint> & points);

#endif //JACOBIAN_H
