#ifndef TIME_CHART_H
#define TIME_CHART_H

#include <string>

class TimeChart {
public:

  struct TimeInfo {
    std::string name;
    double lastTime;
    double totalTime;
    int n;

    TimeInfo(const std::string & _name): name(_name), lastTime(0), totalTime(0), n(0) { }

    void operator<<(Timer & timer) {
      lastTime = double(timer);
      totalTime += lastTime;
      n ++;
    }
  };

  struct Function {
    std::string name;

    Function(const std::string & _name): name(_name) {
      TimeChart::instance().begin(name);
    }

    ~Function() {
      TimeChart::instance().end(name);
    }
  };

  
  std::map<std::string, Timer> beginTimes;
  std::map<std::string, TimeInfo * > timingMap;
  std::vector<TimeInfo * > timings;
  Timer time;

  TimeChart() {
    time.start();
  }

  void begin(std::string name) {
    beginTimes[name] = Timer();
    beginTimes[name].start();
  }

  TimeInfo * getInfo(const std::string & name) {
    if (timingMap.count(name))
      return timingMap[name];
    TimeInfo * info = new TimeInfo(name);
    timingMap[name] = info;
    timings.push_back(info);
    return info;
  }

  void end(const std::string & name) {
    *getInfo(name) << beginTimes[name];
    beginTimes.erase(name);
  }

  void print() {
    double total = 0;

    for (int i=0; i<timings.size(); i++) {
      printf("%12s  %10.4lf %10.3lf %10d\n", timings[i]->name.c_str(),
	     timings[i]->lastTime, timings[i]->totalTime, timings[i]->n);

      total += timings[i]->totalTime;
    }    
    printf("-------------------\n%12s   %10.3lf\n", "total", total);
  }

  static TimeChart & instance() {
    static TimeChart chart;
    return chart;
  }
};


#endif //TIME_CHART_H
