/**
   The main rigid tracker.   Both implementations: image-based jacobian, and texture-based
   factored jacobian.

   Neil Birkbeck, 2010
*/

#include <vector>
#include <map>

#include <utigl/glcommon.h>
#include <utigl/drawcamera.h>

#include <nimage/image.h>
#include <ik/mesh.h>
#include <stereo/simplesequence.h>
#include <autorecon/recon_globals.h>
#include <nmisc/timer.h>
#include <set>
#include <utigl/fbo.h>
#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

#include "boundingbox.h"
#include "utils.h"
#include "jacobian.h"
#include "timechart.h"

#include <vector>
#include <nmath/matrix.h>
#include <ik/mesh.h>

using namespace std;
using namespace nacb;

class Tracker {
public:
  static const float kUpdateThreshold = 1e-5;
  static const float kVisibilityTolerance = 3e-3;
  static const int kUpdateVisibilityEvery = 2;
  vector<seq_t>  seqs;
  vector<GLuint> imageTextures;

  Mesh  mesh;
  Matrix parms;
  Imagef texture;
  GLuint texId;

  std::vector<SamplePoint> points;
  std::vector<Imagef> Tgrad;
  std::vector<std::vector<Imagef> > image_grad;

  std::vector<std::vector<int> > visible;
  
  int nchannels;
  int imagei;

  FrameBufferObject * fbo;
  BoundingBox bounds;

  bool drawProjectedMesh;
  bool drawImageTextures;

  Tracker(std::vector<std::string> seqNames, 
	  const char * meshName, const char * texName);

  ~Tracker();

  std::vector<std::vector<std::pair<int, int> > > updateVisibility();

  void draw();
  void imageTrack(int its = 30);
  void textureTrack(int its = 20);

  void loadImages();
  void computeImageGradients();
  bool trackerLoopIteration(bool imageBased = false);
};
